/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is bookmarktools.
 *
 * The Initial Developer of the Original Code is
 * Maik Gollnick <maik.gollnick@gmx.de>.
 * Portions created by the Initial Developer are Copyright (C) 2008
 * the Initial Developer. All Rights Reserved.
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */

try {
 
const HEADER = Components.classes["@mozilla.org/moz/jssubscript-loader;1"].getService(Components.interfaces.mozIJSSubScriptLoader);
const INCLUDE = function(name){
	try {
		HEADER.loadSubScript("chrome://bookmarktools/content/encoding/" + name);
	} catch(e){}
};
 
INCLUDE("bookmarkToolsASCIIHex.js");
INCLUDE("bookmarkToolsASCII85.js");
INCLUDE("bookmarkToolsDeflate.js");
 
var Encoding = {

	/*
		Returns true if the given mimetype is supported, otherwise false.
		@pre typeof mimeType != 'undefined' && mimeType != null && mimeType.toLowerCase()
	*/
	supportedMimeType : function(mimeType){
		switch(mimeType){
			case "application/rtf" : return true;
			case "application/xhtml+xml" : return true;
			case "application/xml" : return true;
			case "application/x-javascript" : return true;
			case "application/rss+xml" : return true;
			case "text/comma-separated-values" : return true;
			case "text/css" : return true;
			case "text/html" : return true;
			case "text/javascript" : return true;
			case "text/plain" : return true;
			case "text/rtf" : return true;
			case "text/tab-separated-values" : return true;
			case "text/xml" : return true;
			default : return false;
		}
	},

	/*
		Returns true if the given charset is supported, otherwise false.
		@pre typeof charSet != 'undefined' && charSet != null && charSet.toLowerCase()
	*/
	supportedCharSet : function(charSet){
		switch(charSet){
			case "charset=us-ascii" : return true;
			case "charset=utf-8" : return true;
			case "charset=iso-8859-1" : return true;
			case "charset=iso-8859-2" : return true;
			case "charset=iso-8859-3" : return true;
			case "charset=iso-8859-4" : return true;
			default : return false;
		}
	},
	
	isRTF : function(mimeType){
		return mimeType == "application/rtf" || mimeType == "text/rtf";
	},
	
	isHTML : function(mimeType){
		return mimeType == "text/html";
	},
	
	isXML : function(mimeType){
		return mimeType == "application/xhtml+xml" || 
			mimeType == "application/xml" ||
			mimeType == "application/rss+xml" ||
			mimeType == "text/xml";
	},
	
	isPLAIN : function(mimeType){
		return mimeType == "application/x-javascript" || 
			mimeType == "text/comma-separated-values" ||
			mimeType == "text/css" ||
			mimeType == "text/javascript" ||
			mimeType == "text/plain" ||
			mimeType == "text/tab-separated-values";
	},
	
	/*
		Charset independend entity filter.
	*/
	maskedEntity : function(entity){
		switch(entity){
			case "&Agrave;" : // 
			case "&Aacute;" : // 
			case "&Acirc;" : // 
			case "&Atilde;" : // 
			case "&Auml;" : // 
			case "&Aring;" : // 
				return 'A';
			case "&AElig;" : // 
				return "AE";
			case "&Ccedil;" : // 
				return 'C';
			case "&Egrave;" : // 
			case "&Eacute;" : // 
			case "&Ecirc;" : // 
			case "&Euml;" : // 
				return 'E';
			case "&Igrave;" : // 
			case "&Iacute;" : // 
			case "&Icirc;" : // 
			case "&Iuml;" : // 
				return 'I';
			case "&ETH;" : // 
				return 'D';
			case "&Ntilde;" : // 
				return 'N';
			case "&Ograve;" : // 
			case "&Oacute;" : // 
			case "&Ocirc;" : // 
			case "&Otilde;" : // 
			case "&Ouml;" : // 
			case "&Oslash;" : // 
				return 'O';
			case "&THORN;" : // 
				return "TH";
			case "&Ugrave;" : // 
			case "&Uacute;" : // 
			case "&Ucirc;" : // 
			case "&Uuml;" : // 
				return 'U';
			case "&Yacute;" : // 
				return 'Y';
			case "&agrave;" : // 
			case "&aacute;" : // 
			case "&acirc;" : // 
			case "&atilde;" : // 
			case "&auml;" : // 
			case "&aring;" : // 
				return 'a';
			case "&aelig;" : // 
				return "ae";
			case "&ccedil;" : // 
				return 'c';
			case "&egrave;" : // 
			case "&eacute;" : // 
			case "&ecirc;" : // 
			case "&euml;" : // 
				return 'e';
			case "&igrave;" : // 
			case "&iacute;" : // 
			case "&icirc;" : // 
			case "&iuml;" : // 
				return 'i';
			case "&eth;" : // 
				return 'd';
			case "&ntilde;" : // 
				return 'n';
			case "&ograve;" : // 
			case "&oacute;" : // 
			case "&ocirc;" : // 
			case "&otilde;" : // 
			case "&ouml;" : // 
			case "&oslash;" : // 
				return 'o';
			case "&szlig;" : // 
				return "ss";
			case "&thorn;" : // 
				return "th";
			case "&ugrave;" : // 
			case "&uacute;" : // 
			case "&ucirc;" : // 
			case "&uuml;" : // 
				return 'u';
			case "&yacute;" : // 
			case "&yuml;" : // 
				return 'y';
			default : break;
		}
		return null;
	},
	
	/*
		Filter for standard entities.
	*/
	standardEntity : function(entity){
		switch(entity){
			case "&quot;" : return '"';
			case "&#34;" : return '"';
			case "&amp;" : return '&';
			case "&#38;" : return '&';
			case "&lt;" : return '<';
			case "&#60;" : return '<';
			case "&gt;" : return '>';
			case "&#62;" : return '>';
			default : break;
		}
		return null;
	},
	
	/*
		Extracts all document entities as array.
	*/
	extractEntities : function(markup){
		var match = markup.match(/&([a-zA-Z0-9#]+);/gi);
		return match ? match : null;
	},
	
	/*
		Filters all known entities with ascii characters, otherwise the unknown
		entity is replaced by an empty string.
	*/
	filterEntities : function(responseText){
		var entities = Encoding.extractEntities(responseText);
		if(entities != null){
			for(var i = 0, l = entities.length ;i < l; i++){
				var entity = entities[i];
				var standardEntity = Encoding.standardEntity(entity);
				var maskedEntity = Encoding.maskedEntity(entity);
				
				if(standardEntity != null || maskedEntity != null){
					var replaceEntity = maskedEntity != null ? maskedEntity : standardEntity;
					responseText = responseText.replace(entity, replaceEntity);
				}else {
					responseText = responseText.replace(entity, StringUtils.EMPTY);
				}
			}
		}
		return responseText;
	}
	
};

var PDFObject = function(){
	
	var supportedFilters = ['FlateDecode', 'ASCIIHexDecode', 'ASCII85Decode'];
	var properties = [];
	var stream = null;
	
	this.parse = function(content){
		var start = content.indexOf("stream");
		var end = content.indexOf("endstream");
		if(start != -1 && end != -1){
			stream = content.substring(start + 6, end);
			
			start = content.indexOf("<<");
			end = content.indexOf(">>");
			
			// force minimum, dictionary should include Length
			if(start != -1 && end != -1){
				var dictionary = content.substring(start + 2, end).split('/');
				for(var i = 0, l = dictionary.length ;i < l; i++){
					if(!StringUtils.isEmpty(dictionary[i])){
						var property = dictionary[i].split(' ');
						var key = property.shift();
						
						// skip image streams
						if(key == 'Image' || key == 'ColorSpace')
							return false;
						
						properties[key] = property;
					}
				}
				return true;
			}
		}
		return false;
	};
	
	var getFilter = function(){
		if(properties['Filter']){
			var cascade = [];
			for(var i = 0, l = supportedFilters.length ;i < l; i++){
				var filter = supportedFilters[i];
				if(properties[filter])
					cascade.push(filter);
			}
			
			if(ArrayUtils.isEmpty(cascade))
				return null;
				
			return cascade;
		}
		return ['Plain'];
	};
	
	var decode = function(byteStream, filter){
		switch(filter){
			case 'FlateDecode':
				return Deflate.decode(byteStream);
			case 'ASCIIHexDecode':
				return ASCIIHex.decode(byteStream);
			case 'ASCII85Decode':
				return ASCII85.decode(byteStream);
			case 'Plain':
				var write = false;
				var out = [];
				for(var i = 0, l = byteStream.length ;i < l; i++){
					if(write)
						out.push(byteStream[i]);
						
					switch(byteStream[i]){
						case 40: //'('
							write = true;
							break;
						case 41: //')'
							out.pop();
							out.push(32);
							write = false;
							break;
						default: break;
					}
				}
				byteStream = out;
				break;
		}
		return byteStream;
	};
	
	this.extractText = function(){
		var cascade = getFilter(), filter;

		if(cascade == null)
			return StringUtils.EMPTY;
			
		var byteStream = StringUtils.bytesOf(stream);
		for(var i = 0, l = cascade.length ;i < l; i++){
			filter = cascade[i];
			byteStream = decode(byteStream, filter);
		}
		return StringUtils.stringOf(byteStream);
	};
	
};

var PDFDocument = {

	version : -1,
	header : null,
	trailer : null,
	objects : null,

	test : function(filename){
		/*var file = Cc["@mozilla.org/file/directory_service;1"].getService(Ci.nsIProperties).get("ProfD", Ci.nsIFile);
		file.append(filename);
		var content = BookmarkTools.utils.fileio.readFile(file);
		
		this.parse(content);*/
		try {
			var instDeflater = Cc["@bookmarktools.mozdev.org/encoding/bookmarkToolsZlib;1"].createInstance();
			var deflater = instDeflater.QueryInterface(Ci.BookmarkToolsDeflater);
			
			var src = "aaaaaaaaaaaaaaaabbbbbbbbbbbbbbbbbbbbbbbb";
			var dfl = deflater.deflate(src);
			var ifl = deflater.inflate(dfl);
			
			alert(src + " (" + src.length + ")\n" + dfl+ " (" + dfl.length + ")\n"+ifl);
			
		}catch(e){
			alert(e);
		}
	},
	
	parse : function(content){
		content = content.replace(/[\n\r]*/gi, StringUtils.EMPTY);
		var pos = 0;
		
		var obj, start, end;
		var objects = [];

		do {
			start = content.indexOf("obj", pos);
			end = content.indexOf("endobj", pos);
			
			if(start != -1 && end != -1){
				obj = content.substring(start, end + 6);
				
				var pdfObj = new PDFObject();
				if(pdfObj.parse(obj)){
					objects.push(pdfObj);
					alert(pdfObj.extractText());
				}
				
				pos = end + 6;
			}
		}while(start != -1 && end != -1);
	},
	
	getVersion : function(){},
	getHeader : function(){},
	getTrailer : function(){}
	
};

//PDFDocument.test("test.pdf");

var Parser = {

	tidy : null,

	init : function(){
		var instTidy = Cc["@bookmarktools.mozdev.org/repair/tidy;1"].createInstance();
		Parser.tidy = instTidy.QueryInterface(Ci.ITidy);
	},
	
	_init : function(){
		Parser.tidy = null;
	},

	removeWhiteSpaces : function(content){
		return content.replace(/(\s)+/gi, StringUtils.EMPTY);
	},
	
	removeNumbers : function(content){
		return content.replace(/[0-9]+/gi, StringUtils.EMPTY);
	},
	
	replaceWhiteSpaces : function(content){
		return content.replace(/(\s)+/gi, ' ');
	},
	
	parseRTF : function(responseText){
		responseText = responseText.replace(/\{?\\[a-zA-Z0-9*-]+\s*/gi, StringUtils.EMPTY);
		responseText = responseText.replace(/\s*[a-zA-Z\s]+;\}/gi, StringUtils.EMPTY);
		return responseText.replace(/\}/gi, StringUtils.EMPTY);
	},

	parse : function(mimeType, charSet, responseText){
		if(Encoding.isPLAIN(mimeType)){
			return responseText;
		}else {
			if(Encoding.isRTF(mimeType)){
				return Parser.parseRTF(responseText);
			}else {
				if(Encoding.isHTML(mimeType) || Encoding.isXML(mimeType)){
					var response = Parser.tidy.cleanup(responseText);
					if(response != "Tidy: no output"){
						response = Parser.replaceWhiteSpaces(response);
						response = StringUtils.stripScripts(response); 
						response = StringUtils.stripTags(response); 
						response = Encoding.filterEntities(response);
						return response;
					}else {
						return responseText;
					}
				}else {
					throw "Error(Parser): Unsupported mimetype: " + mimeType;
				}
			}
		}
	}
	
}; 

} catch(e){
	alert(e);
}